@@ -29,18 +29,24 @@ class TftpServer(TftpSession):
         # ip:tid for the remote end.
         self.sessions = {}
 
-        if os.path.exists(self.root):
-            log.debug("tftproot %s does exist" % self.root)
+        self.shutdown_gracefully = False
+        self.shutdown_immediately = False
+
+        if self.dyn_file_func:
+            if not callable(self.dyn_file_func):
+                raise TftpException, "A dyn_file_func supplied, but it is not callable."
+        elif os.path.exists(self.root):
+            log.debug("tftproot %s does exist", self.root)
             if not os.path.isdir(self.root):
                 raise TftpException, "The tftproot must be a directory."
             else:
-                log.debug("tftproot %s is a directory" % self.root)
+                log.debug("tftproot %s is a directory", self.root)
                 if os.access(self.root, os.R_OK):
-                    log.debug("tftproot %s is readable" % self.root)
+                    log.debug("tftproot %s is readable", self.root)
                 else:
                     raise TftpException, "The tftproot must be readable"
                 if os.access(self.root, os.W_OK):
-                    log.debug("tftproot %s is writable" % self.root)
+                    log.debug("tftproot %s is writable", self.root)
                 else:
                     log.warning("The tftproot %s is not writable" % self.root)
         else:
@@ -70,6 +76,20 @@ class TftpServer(TftpSession):
 
         log.info("Starting receive loop...")
         while True:
+            if self.shutdown_immediately:
+                log.warn("Shutting down now. Session count: %d" % len(self.sessions))
+                self.sock.close()
+                for key in self.sessions:
+                    self.sessions[key].end()
+                self.sessions = []
+                break
+
+            elif self.shutdown_gracefully:
+                if not self.sessions:
+                    log.warn("In graceful shutdown mode and all sessions complete.")
+                    self.sock.close()
+                    break
+
             # Build the inputlist array of sockets to select() on.
             inputlist = []
             inputlist.append(self.sock)
@@ -77,7 +97,7 @@ class TftpServer(TftpSession):
                 inputlist.append(self.sessions[key].sock)
 
             # Block until some socket has input on it.
-            log.debug("Performing select on this inputlist: %s" % inputlist)
+            log.debug("Performing select on this inputlist: %s", inputlist)
             readyinput, readyoutput, readyspecial = select.select(inputlist,
                                                                   [],
                                                                   [],
@@ -92,7 +112,11 @@ class TftpServer(TftpSession):
                     log.debug("Data ready on our main socket")
                     buffer, (raddress, rport) = self.sock.recvfrom(MAX_BLKSIZE)
 
-                    log.debug("Read %d bytes" % len(buffer))
+                    log.debug("Read %d bytes", len(buffer))
+
+                    if self.shutdown_gracefully:
+                        log.warn("Discarding data on main port, in graceful shutdown mode")
+                        continue
 
                     # Forge a session key based on the client's IP and port,
                     # which should safely work through NAT.
@@ -100,7 +124,7 @@ class TftpServer(TftpSession):
 
                     if not self.sessions.has_key(key):
                         log.debug("Creating new server context for "
-                                     "session key = %s" % key)
+                                     "session key = %s", key)
                         self.sessions[key] = TftpContextServer(raddress,
                                                                rport,
                                                                timeout,
@@ -152,12 +176,11 @@ class TftpServer(TftpSession):
                     log.error(str(err))
                     self.sessions[key].retry_count += 1
                     if self.sessions[key].retry_count >= TIMEOUT_RETRIES:
-                        log.debug("hit max retries on %s, giving up"
-                            % self.sessions[key])
+                        log.debug("hit max retries on %s, giving up",
+                            self.sessions[key])
                         deletion_list.append(key)
                     else:
-                        log.debug("resending on session %s"
-                            % self.sessions[key])
+                        log.debug("resending on session %s", self.sessions[key])
                         self.sessions[key].state.resendLast()
 
             log.debug("Iterating deletion list.")
@@ -176,9 +199,25 @@ class TftpServer(TftpSession):
                         log.info("Average rate: %.2f kbps" % metrics.kbps)
                     log.info("%.2f bytes in resent data" % metrics.resent_bytes)
                     log.info("%d duplicate packets" % metrics.dupcount)
-                    log.debug("Deleting session %s" % key)
+                    log.debug("Deleting session %s", key)
                     del self.sessions[key]
-                    log.debug("Session list is now %s" % self.sessions)
+                    log.debug("Session list is now %s", self.sessions)
                 else:
                     log.warn("Strange, session %s is not on the deletion list"
                         % key)
+
+        log.debug("server returning from while loop")
+        self.shutdown_gracefully = self.shutdown_immediately = False
+
+    def stop(force=False):
+        """Stop the server gracefully. Do not take any new transfers,
+        but complete the existing ones. If force is True, drop everything
+        and stop. Note, immediately will not interrupt the select loop, it
+        will happen when the server returns on ready data, or a timeout.
+        ie. SOCK_TIMEOUT"""
+        if force:
+            log.info("Server instructed to shut down immediately.")
+            self.shutdown_immediately = True
+        else:
+            log.info("Server instructed to shut down gracefully.")
+            self.shutdown_gracefully = True
